from enum import Enum
from typing import Any, Dict

from pydantic import BaseModel, Field, model_validator


class MetricType(str, Enum):
    """
    Metric Constant for milvus/ zilliz server.
    """

    def __str__(self) -> str:
        return str(self.value)

    L2 = "L2"
    IP = "IP"
    COSINE = "COSINE"
    HAMMING = "HAMMING"
    JACCARD = "JACCARD"


class MilvusDBConfig(BaseModel):
    url: str = Field("http://localhost:19530", description="Full URL for Milvus/Zilliz server")
    token: str = Field(None, description="Token for Zilliz server / local setup defaults to None.")
    collection_name: str = Field("mem0", description="Name of the collection")
    embedding_model_dims: int = Field(1536, description="Dimensions of the embedding model")
    metric_type: str = Field("L2", description="Metric type for similarity search")

    @model_validator(mode="before")
    @classmethod
    def validate_extra_fields(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        allowed_fields = set(cls.model_fields.keys())
        input_fields = set(values.keys())
        extra_fields = input_fields - allowed_fields
        if extra_fields:
            raise ValueError(
                f"Extra fields not allowed: {', '.join(extra_fields)}. Please input only the following fields: {', '.join(allowed_fields)}"
            )
        return values

    model_config = {
        "arbitrary_types_allowed": True,
    }



"""
Pydantic models for standardizing LLM outputs in Mem0's graph memory system.
These models ensure structured and predictable outputs from LLMs, especially
when using small local models with Ollama's format argument.
"""

from typing import List, Optional
from pydantic import BaseModel, Field


class Entity(BaseModel):
    """Model representing an entity extracted from text."""
    entity: str = Field(..., description="The name or identifier of the entity")
    entity_type: str = Field(..., description="The type or category of the entity")


class Relationship(BaseModel):
    """Model representing a relationship between two entities."""
    source: str = Field(..., description="The source entity of the relationship")
    relationship: str = Field(..., description="The relationship between the source and destination entities")
    destination: str = Field(..., description="The destination entity of the relationship")


class EntityExtraction(BaseModel):
    """Model representing the output of entity extraction."""
    entities: List[Entity] = Field(..., description="List of entities extracted from text")


class RelationshipExtraction(BaseModel):
    """Model representing the output of relationship extraction."""
    entities: List[Relationship] = Field(..., description="List of relationships between entities")


class DeleteMemory(BaseModel):
    """Model representing a memory to be deleted."""
    source: str = Field(..., description="The source entity in the relationship")
    relationship: str = Field(..., description="The relationship between entities")
    destination: str = Field(..., description="The destination entity in the relationship")


class DeleteMemoryOutput(BaseModel):
    """Model representing the output of memory deletion."""
    entities: List[DeleteMemory] = Field(default_factory=list, description="List of memories to delete")


class UpdateMemory(BaseModel):
    """Model representing a memory to be updated."""
    source: str = Field(..., description="The source entity in the relationship")
    relationship: str = Field(..., description="The relationship between entities")
    destination: str = Field(..., description="The destination entity in the relationship")


class UpdateMemoryOutput(BaseModel):
    """Model representing the output of memory updates."""
    entities: List[UpdateMemory] = Field(default_factory=list, description="List of memories to update")


class AddMemory(BaseModel):
    """Model representing a memory to be added."""
    source: str = Field(..., description="The source entity in the relationship")
    destination: str = Field(..., description="The destination entity in the relationship")
    relationship: str = Field(..., description="The relationship between entities")
    source_type: str = Field(..., description="The type of the source entity")
    destination_type: str = Field(..., description="The type of the destination entity")


class AddMemoryOutput(BaseModel):
    """Model representing the output of memory addition."""
    entities: List[AddMemory] = Field(default_factory=list, description="List of memories to add")


class NoopOutput(BaseModel):
    """Model representing a no-operation output."""
    message: str = Field(default="No operation needed", description="Message indicating no operation is needed")



from typing import Optional

from pydantic import BaseModel, Field, field_validator, model_validator

from mem0llama.llms.configs import LlmConfig


class Neo4jConfig(BaseModel):
    url: Optional[str] = Field(None, description="Host address for the graph database")
    username: Optional[str] = Field(None, description="Username for the graph database")
    password: Optional[str] = Field(None, description="Password for the graph database")

    @model_validator(mode="before")
    def check_host_port_or_path(cls, values):
        url, username, password = (
            values.get("url"),
            values.get("username"),
            values.get("password"),
        )
        if not url or not username or not password:
            raise ValueError("Please provide 'url', 'username' and 'password'.")
        return values


class GraphStoreConfig(BaseModel):
    provider: str = Field(description="Provider of the data store (e.g., 'neo4j')", default="neo4j")
    config: Neo4jConfig = Field(description="Configuration for the specific data store", default=None)
    llm: Optional[LlmConfig] = Field(description="LLM configuration for querying the graph store", default=None)
    custom_prompt: Optional[str] = Field(
        description="Custom prompt to fetch entities from the given text", default=None
    )

    @field_validator("config")
    def validate_config(cls, v, values):
        provider = values.data.get("provider")
        if provider == "neo4j":
            return Neo4jConfig(**v.model_dump())
        else:
            raise ValueError(f"Unsupported graph store provider: {provider}")



import logging

from mem0llama.memory.utils import format_entities
from mem0llama.memory.models import (
    EntityExtraction,
    RelationshipExtraction,
    DeleteMemoryOutput,
    UpdateMemoryOutput,
    AddMemoryOutput,
    NoopOutput
)
from mem0llama.memory.llm_formatter import (
    parse_llm_response,
    create_entity_extraction_prompt,
    create_relationship_extraction_prompt,
    create_delete_memory_prompt,
    adapt_llm_provider_for_structured_output
)

try:
    from langchain_neo4j import Neo4jGraph
except ImportError:
    raise ImportError("langchain_community is not installed. Please install it using pip install langchain-community")

try:
    from rank_bm25 import BM25Okapi
except ImportError:
    raise ImportError("rank_bm25 is not installed. Please install it using pip install rank-bm25")

from mem0llama.graphs.tools import (
    DELETE_MEMORY_STRUCT_TOOL_GRAPH,
    DELETE_MEMORY_TOOL_GRAPH,
    EXTRACT_ENTITIES_STRUCT_TOOL,
    EXTRACT_ENTITIES_TOOL,
    RELATIONS_STRUCT_TOOL,
    RELATIONS_TOOL,
)
from mem0llama.graphs.utils import EXTRACT_RELATIONS_PROMPT, get_delete_messages
from mem0llama.utils.factory import EmbedderFactory, LlmFactory

logger = logging.getLogger(__name__)


class MemoryGraph:
    def __init__(self, config):
        self.config = config
        self.graph = Neo4jGraph(
            self.config.graph_store.config.url,
            self.config.graph_store.config.username,
            self.config.graph_store.config.password,
        )
        self.embedding_model = EmbedderFactory.create(self.config.embedder.provider, self.config.embedder.config)

        self.llm_provider = "openai_structured"
        if self.config.llm.provider:
            self.llm_provider = self.config.llm.provider
        if self.config.graph_store.llm:
            self.llm_provider = self.config.graph_store.llm.provider

        # Adapt LLM config for structured output
        llm_config = self.config.llm.config
        if self.config.graph_store.llm:
            llm_config = self.config.graph_store.llm.config
        
        self.adapted_llm_config = adapt_llm_provider_for_structured_output(
            self.llm_provider, llm_config
        )
        
        self.llm = LlmFactory.create(self.llm_provider, self.adapted_llm_config)
        self.user_id = None
        self.threshold = 0.7

    def add(self, data, filters):
        """
        Adds data to the graph.

        Args:
            data (str): The data to add to the graph.
            filters (dict): A dictionary containing filters to be applied during the addition.
        """
        entity_type_map = self._retrieve_nodes_from_data(data, filters)
        to_be_added = self._establish_nodes_relations_from_data(data, filters, entity_type_map)
        search_output = self._search_graph_db(node_list=list(entity_type_map.keys()), filters=filters)
        to_be_deleted = self._get_delete_entities_from_search_output(search_output, data, filters)

        # TODO: Batch queries with APOC plugin
        # TODO: Add more filter support
        deleted_entities = self._delete_entities(to_be_deleted, filters["user_id"])
        added_entities = self._add_entities(to_be_added, filters["user_id"], entity_type_map)

        return {"deleted_entities": deleted_entities, "added_entities": added_entities}

    def search(self, query, filters, limit=100):
        """
        Search for memories and related graph data.

        Args:
            query (str): Query to search for.
            filters (dict): A dictionary containing filters to be applied during the search.
            limit (int): The maximum number of nodes and relationships to retrieve. Defaults to 100.

        Returns:
            dict: A dictionary containing:
                - "contexts": List of search results from the base data store.
                - "entities": List of related graph data based on the query.
        """
        entity_type_map = self._retrieve_nodes_from_data(query, filters)
        search_output = self._search_graph_db(node_list=list(entity_type_map.keys()), filters=filters)

        if not search_output:
            return []

        search_outputs_sequence = [
            [item["source"], item["relationship"], item["destination"]] for item in search_output
        ]
        bm25 = BM25Okapi(search_outputs_sequence)

        tokenized_query = query.split(" ")
        reranked_results = bm25.get_top_n(tokenized_query, search_outputs_sequence, n=5)

        search_results = []
        for item in reranked_results:
            search_results.append({"source": item[0], "relationship": item[1], "destination": item[2]})

        logger.info(f"Returned {len(search_results)} search results")

        return search_results

    def delete_all(self, filters):
        cypher = """
        MATCH (n {user_id: $user_id})
        DETACH DELETE n
        """
        params = {"user_id": filters["user_id"]}
        self.graph.query(cypher, params=params)

    def get_all(self, filters, limit=100):
        """
        Retrieves all nodes and relationships from the graph database based on optional filtering criteria.

        Args:
            filters (dict): A dictionary containing filters to be applied during the retrieval.
            limit (int): The maximum number of nodes and relationships to retrieve. Defaults to 100.
        Returns:
            list: A list of dictionaries, each containing:
                - 'contexts': The base data store response for each memory.
                - 'entities': A list of strings representing the nodes and relationships
        """

        # return all nodes and relationships
        query = """
        MATCH (n {user_id: $user_id})-[r]->(m {user_id: $user_id})
        RETURN n.name AS source, type(r) AS relationship, m.name AS target
        LIMIT $limit
        """
        results = self.graph.query(query, params={"user_id": filters["user_id"], "limit": limit})

        final_results = []
        for result in results:
            final_results.append(
                {
                    "source": result["source"],
                    "relationship": result["relationship"],
                    "target": result["target"],
                }
            )

        logger.info(f"Retrieved {len(final_results)} relationships")

        return final_results

    def _retrieve_nodes_from_data(self, data, filters):
        """Extracts all the entities mentioned in the query."""
        entity_type_map = {}
        
        # Create a structured prompt for entity extraction
        prompt = create_entity_extraction_prompt(data, filters["user_id"])
        
        # Determine if we should use tools based on the LLM provider
        use_tools = self.llm_provider in ["azure_openai_structured", "openai_structured"]
        
        if use_tools:
            _tools = [EXTRACT_ENTITIES_STRUCT_TOOL]
            search_results = self.llm.generate_response(
                messages=[
                    {
                        "role": "system",
                        "content": f"You are a smart assistant who understands entities and their types in a given text. If user message contains self reference such as 'I', 'me', 'my' etc. then use {filters['user_id']} as the source entity. Extract all the entities from the text. ***DO NOT*** answer the question itself if the given text is a question.",
                    },
                    {"role": "user", "content": data},
                ],
                tools=_tools,
            )
            
            try:
                for item in search_results["tool_calls"][0]["arguments"]["entities"]:
                    entity_type_map[item["entity"]] = item["entity_type"]
            except Exception as e:
                logger.error(f"Error in search tool: {e}")
        else:
            # For local LLMs, use direct prompting with structured output
            response = self.llm.generate_response(
                messages=[
                    {"role": "system", "content": "You are an AI assistant that extracts entities and their types from text."},
                    {"role": "user", "content": prompt}
                ],
                response_format=EntityExtraction.model_json_schema()
            )
            
            # Parse the response using our Pydantic model
            try:
                # For Ollama with format parameter, the response might be directly parseable
                if isinstance(response, dict) and "message" in response and "content" in response["message"]:
                    # Direct Ollama response
                    extraction_result = EntityExtraction.model_validate_json(response["message"]["content"])
                elif isinstance(response, dict) and "content" in response:
                    # Standard response format
                    extraction_result = parse_llm_response(response["content"], EntityExtraction)
                else:
                    # Fallback to our parser
                    extraction_result = parse_llm_response(response, EntityExtraction)
                
                for entity in extraction_result.entities:
                    entity_type_map[entity.entity] = entity.entity_type
            except Exception as e:
                logger.error(f"Error parsing entity extraction response: {e}")
                logger.debug(f"Raw response: {response}")

        entity_type_map = {k.lower().replace(" ", "_"): v.lower().replace(" ", "_") for k, v in entity_type_map.items()}
        logger.debug(f"Entity type map: {entity_type_map}")
        return entity_type_map

    def _establish_nodes_relations_from_data(self, data, filters, entity_type_map):
        """Establish relations among the extracted nodes."""
        # Create a structured prompt for relationship extraction
        entity_list = list(entity_type_map.keys())
        prompt = create_relationship_extraction_prompt(data, entity_list, filters["user_id"])
        
        # Determine if we should use tools based on the LLM provider
        use_tools = self.llm_provider in ["azure_openai_structured", "openai_structured"]
        
        if use_tools:
            if self.config.graph_store.custom_prompt:
                messages = [
                    {
                        "role": "system",
                        "content": EXTRACT_RELATIONS_PROMPT.replace("USER_ID", filters["user_id"]).replace(
                            "CUSTOM_PROMPT", f"4. {self.config.graph_store.custom_prompt}"
                        ),
                    },
                    {"role": "user", "content": data},
                ]
            else:
                messages = [
                    {
                        "role": "system",
                        "content": EXTRACT_RELATIONS_PROMPT.replace("USER_ID", filters["user_id"]),
                    },
                    {"role": "user", "content": f"List of entities: {list(entity_type_map.keys())}. \n\nText: {data}"},
                ]

            _tools = [RELATIONS_STRUCT_TOOL]
            extracted_entities_response = self.llm.generate_response(
                messages=messages,
                tools=_tools,
            )

            if extracted_entities_response["tool_calls"]:
                extracted_entities = extracted_entities_response["tool_calls"][0]["arguments"]["entities"]
            else:
                extracted_entities = []
        else:
            # For local LLMs, use direct prompting with structured output
            response = self.llm.generate_response(
                messages=[
                    {"role": "system", "content": "You are an AI assistant that identifies relationships between entities."},
                    {"role": "user", "content": prompt}
                ],
                response_format=RelationshipExtraction.model_json_schema()
            )
            
            # Parse the response using our Pydantic model
            try:
                # For Ollama with format parameter, the response might be directly parseable
                if isinstance(response, dict) and "message" in response and "content" in response["message"]:
                    # Direct Ollama response
                    extraction_result = RelationshipExtraction.model_validate_json(response["message"]["content"])
                elif isinstance(response, dict) and "content" in response:
                    # Standard response format
                    extraction_result = parse_llm_response(response["content"], RelationshipExtraction)
                else:
                    # Fallback to our parser
                    extraction_result = parse_llm_response(response, RelationshipExtraction)
                
                # Convert to the format expected by the rest of the code
                extracted_entities = []
                for relationship in extraction_result.entities:
                    extracted_entities.append({
                        "source": relationship.source,
                        "relationship": relationship.relationship,
                        "destination": relationship.destination
                    })
            except Exception as e:
                logger.error(f"Error parsing relationship extraction response: {e}")
                logger.debug(f"Raw response: {response}")
                extracted_entities = []

        extracted_entities = self._remove_spaces_from_entities(extracted_entities)
        logger.debug(f"Extracted entities: {extracted_entities}")
        return extracted_entities

    def _search_graph_db(self, node_list, filters, limit=100):
        """Search similar nodes among and their respective incoming and outgoing relations."""
        result_relations = []

        for node in node_list:
            n_embedding = self.embedding_model.embed(node)

            cypher_query = """
            MATCH (n)
            WHERE n.embedding IS NOT NULL AND n.user_id = $user_id
            WITH n,
                round(reduce(dot = 0.0, i IN range(0, size(n.embedding)-1) | dot + n.embedding[i] * $n_embedding[i]) /
                (sqrt(reduce(l2 = 0.0, i IN range(0, size(n.embedding)-1) | l2 + n.embedding[i] * n.embedding[i])) *
                sqrt(reduce(l2 = 0.0, i IN range(0, size($n_embedding)-1) | l2 + $n_embedding[i] * $n_embedding[i]))), 4) AS similarity
            WHERE similarity >= $threshold
            MATCH (n)-[r]->(m)
            RETURN n.name AS source, elementId(n) AS source_id, type(r) AS relationship, elementId(r) AS relation_id, m.name AS destination, elementId(m) AS destination_id, similarity
            UNION
            MATCH (n)
            WHERE n.embedding IS NOT NULL AND n.user_id = $user_id
            WITH n,
                round(reduce(dot = 0.0, i IN range(0, size(n.embedding)-1) | dot + n.embedding[i] * $n_embedding[i]) /
                (sqrt(reduce(l2 = 0.0, i IN range(0, size(n.embedding)-1) | l2 + n.embedding[i] * n.embedding[i])) *
                sqrt(reduce(l2 = 0.0, i IN range(0, size($n_embedding)-1) | l2 + $n_embedding[i] * $n_embedding[i]))), 4) AS similarity
            WHERE similarity >= $threshold
            MATCH (m)-[r]->(n)
            RETURN m.name AS source, elementId(m) AS source_id, type(r) AS relationship, elementId(r) AS relation_id, n.name AS destination, elementId(n) AS destination_id, similarity
            ORDER BY similarity DESC
            LIMIT $limit
            """
            params = {
                "n_embedding": n_embedding,
                "threshold": self.threshold,
                "user_id": filters["user_id"],
                "limit": limit,
            }
            ans = self.graph.query(cypher_query, params=params)
            result_relations.extend(ans)

        return result_relations

    def _get_delete_entities_from_search_output(self, search_output, data, filters):
        """Get the entities to be deleted from the search output."""
        search_output_string = format_entities(search_output)
        
        # Create a structured prompt for memory deletion
        prompt = create_delete_memory_prompt(search_output_string, data, filters["user_id"])
        
        # Determine if we should use tools based on the LLM provider
        use_tools = self.llm_provider in ["azure_openai_structured", "openai_structured"]
        
        if use_tools:
            system_prompt, user_prompt = get_delete_messages(search_output_string, data, filters["user_id"])
            _tools = [DELETE_MEMORY_STRUCT_TOOL_GRAPH]
            
            memory_updates = self.llm.generate_response(
                messages=[
                    {"role": "system", "content": system_prompt},
                    {"role": "user", "content": user_prompt},
                ],
                tools=_tools,
            )
            
            to_be_deleted = []
            for item in memory_updates["tool_calls"]:
                if item["name"] == "delete_graph_memory":
                    to_be_deleted.append(item["arguments"])
        else:
            # For local LLMs, use direct prompting with structured output
            response = self.llm.generate_response(
                messages=[
                    {"role": "system", "content": "You are an AI assistant that identifies relationships to be deleted from a knowledge graph."},
                    {"role": "user", "content": prompt}
                ],
                response_format=DeleteMemoryOutput.model_json_schema()
            )
            
            # Parse the response using our Pydantic model
            try:
                # For Ollama with format parameter, the response might be directly parseable
                if isinstance(response, dict) and "message" in response and "content" in response["message"]:
                    # Direct Ollama response
                    deletion_result = DeleteMemoryOutput.model_validate_json(response["message"]["content"])
                elif isinstance(response, dict) and "content" in response:
                    # Standard response format
                    deletion_result = parse_llm_response(response["content"], DeleteMemoryOutput)
                else:
                    # Fallback to our parser
                    deletion_result = parse_llm_response(response, DeleteMemoryOutput)
                
                # Convert to the format expected by the rest of the code
                to_be_deleted = []
                for memory in deletion_result.entities:
                    to_be_deleted.append({
                        "source": memory.source,
                        "relationship": memory.relationship,
                        "destination": memory.destination
                    })
            except Exception as e:
                logger.error(f"Error parsing memory deletion response: {e}")
                logger.debug(f"Raw response: {response}")
                to_be_deleted = []
                
        # in case if it is not in the correct format
        to_be_deleted = self._remove_spaces_from_entities(to_be_deleted)
        logger.debug(f"Deleted relationships: {to_be_deleted}")
        return to_be_deleted

    def _delete_entities(self, to_be_deleted, user_id):
        """Delete the entities from the graph."""
        results = []
        for item in to_be_deleted:
            source = item["source"]
            destination = item["destination"]
            relationship = item["relationship"]

            # Delete the specific relationship between nodes
            cypher = f"""
            MATCH (n {{name: $source_name, user_id: $user_id}})
            -[r:{relationship}]->
            (m {{name: $dest_name, user_id: $user_id}})
            DELETE r
            RETURN 
                n.name AS source,
                m.name AS target,
                type(r) AS relationship
            """
            params = {
                "source_name": source,
                "dest_name": destination,
                "user_id": user_id,
            }
            result = self.graph.query(cypher, params=params)
            results.append(result)
        return results

    def _add_entities(self, to_be_added, user_id, entity_type_map):
        """Add the new entities to the graph. Merge the nodes if they already exist."""
        results = []
        for item in to_be_added:
            # entities
            source = item["source"]
            destination = item["destination"]
            relationship = item["relationship"]

            # types
            source_type = entity_type_map.get(source, "unknown")
            destination_type = entity_type_map.get(destination, "unknown")

            # embeddings
            source_embedding = self.embedding_model.embed(source)
            dest_embedding = self.embedding_model.embed(destination)

            # search for the nodes with the closest embeddings
            source_node_search_result = self._search_source_node(source_embedding, user_id, threshold=0.9)
            destination_node_search_result = self._search_destination_node(dest_embedding, user_id, threshold=0.9)

            # TODO: Create a cypher query and common params for all the cases
            if not destination_node_search_result and source_node_search_result:
                cypher = f"""
                    MATCH (source)
                    WHERE elementId(source) = $source_id
                    MERGE (destination:{destination_type} {{name: $destination_name, user_id: $user_id}})
                    ON CREATE SET
                        destination.created = timestamp(),
                        destination.embedding = $destination_embedding
                    MERGE (source)-[r:{relationship}]->(destination)
                    ON CREATE SET 
                        r.created = timestamp()
                    RETURN source.name AS source, type(r) AS relationship, destination.name AS target
                    """

                params = {
                    "source_id": source_node_search_result[0]["elementId(source_candidate)"],
                    "destination_name": destination,
                    "relationship": relationship,
                    "destination_type": destination_type,
                    "destination_embedding": dest_embedding,
                    "user_id": user_id,
                }
                resp = self.graph.query(cypher, params=params)
                results.append(resp)

            elif destination_node_search_result and not source_node_search_result:
                cypher = f"""
                    MATCH (destination)
                    WHERE elementId(destination) = $destination_id
                    MERGE (source:{source_type} {{name: $source_name, user_id: $user_id}})
                    ON CREATE SET
                        source.created = timestamp(),
                        source.embedding = $source_embedding
                    MERGE (source)-[r:{relationship}]->(destination)
                    ON CREATE SET 
                        r.created = timestamp()
                    RETURN source.name AS source, type(r) AS relationship, destination.name AS target
                    """

                params = {
                    "destination_id": destination_node_search_result[0]["elementId(destination_candidate)"],
                    "source_name": source,
                    "relationship": relationship,
                    "source_type": source_type,
                    "source_embedding": source_embedding,
                    "user_id": user_id,
                }
                resp = self.graph.query(cypher, params=params)
                results.append(resp)

            elif source_node_search_result and destination_node_search_result:
                cypher = f"""
                    MATCH (source)
                    WHERE elementId(source) = $source_id
                    MATCH (destination)
                    WHERE elementId(destination) = $destination_id
                    MERGE (source)-[r:{relationship}]->(destination)
                    ON CREATE SET 
                        r.created_at = timestamp(),
                        r.updated_at = timestamp()
                    RETURN source.name AS source, type(r) AS relationship, destination.name AS target
                    """
                params = {
                    "source_id": source_node_search_result[0]["elementId(source_candidate)"],
                    "destination_id": destination_node_search_result[0]["elementId(destination_candidate)"],
                    "user_id": user_id,
                    "relationship": relationship,
                }
                resp = self.graph.query(cypher, params=params)
                results.append(resp)

            elif not source_node_search_result and not destination_node_search_result:
                cypher = f"""
                    MERGE (n:{source_type} {{name: $source_name, user_id: $user_id}})
                    ON CREATE SET n.created = timestamp(), n.embedding = $source_embedding
                    ON MATCH SET n.embedding = $source_embedding
                    MERGE (m:{destination_type} {{name: $dest_name, user_id: $user_id}})
                    ON CREATE SET m.created = timestamp(), m.embedding = $dest_embedding
                    ON MATCH SET m.embedding = $dest_embedding
                    MERGE (n)-[rel:{relationship}]->(m)
                    ON CREATE SET rel.created = timestamp()
                    RETURN n.name AS source, type(rel) AS relationship, m.name AS target
                    """
                params = {
                    "source_name": source,
                    "source_type": source_type,
                    "dest_name": destination,
                    "destination_type": destination_type,
                    "source_embedding": source_embedding,
                    "dest_embedding": dest_embedding,
                    "user_id": user_id,
                }
                resp = self.graph.query(cypher, params=params)
                results.append(resp)
        return results

    def _remove_spaces_from_entities(self, entity_list):
        for item in entity_list:
            item["source"] = item["source"].lower().replace(" ", "_")
            item["relationship"] = item["relationship"].lower().replace(" ", "_")
            item["destination"] = item["destination"].lower().replace(" ", "_")
        return entity_list

    def _search_source_node(self, source_embedding, user_id, threshold=0.9):
        cypher = """
            MATCH (source_candidate)
            WHERE source_candidate.embedding IS NOT NULL 
            AND source_candidate.user_id = $user_id

            WITH source_candidate,
                round(
                    reduce(dot = 0.0, i IN range(0, size(source_candidate.embedding)-1) |
                        dot + source_candidate.embedding[i] * $source_embedding[i]) /
                    (sqrt(reduce(l2 = 0.0, i IN range(0, size(source_candidate.embedding)-1) |
                        l2 + source_candidate.embedding[i] * source_candidate.embedding[i])) *
                    sqrt(reduce(l2 = 0.0, i IN range(0, size($source_embedding)-1) |
                        l2 + $source_embedding[i] * $source_embedding[i])))
                , 4) AS source_similarity
            WHERE source_similarity >= $threshold

            WITH source_candidate, source_similarity
            ORDER BY source_similarity DESC
            LIMIT 1

            RETURN elementId(source_candidate)
            """

        params = {
            "source_embedding": source_embedding,
            "user_id": user_id,
            "threshold": threshold,
        }

        result = self.graph.query(cypher, params=params)
        return result

    def _search_destination_node(self, destination_embedding, user_id, threshold=0.9):
        cypher = """
            MATCH (destination_candidate)
            WHERE destination_candidate.embedding IS NOT NULL 
            AND destination_candidate.user_id = $user_id

            WITH destination_candidate,
                round(
                    reduce(dot = 0.0, i IN range(0, size(destination_candidate.embedding)-1) |
                        dot + destination_candidate.embedding[i] * $destination_embedding[i]) /
                    (sqrt(reduce(l2 = 0.0, i IN range(0, size(destination_candidate.embedding)-1) |
                        l2 + destination_candidate.embedding[i] * destination_candidate.embedding[i])) *
                    sqrt(reduce(l2 = 0.0, i IN range(0, size($destination_embedding)-1) |
                        l2 + $destination_embedding[i] * $destination_embedding[i])))
                , 4) AS destination_similarity
            WHERE destination_similarity >= $threshold

            WITH destination_candidate, destination_similarity
            ORDER BY destination_similarity DESC
            LIMIT 1

            RETURN elementId(destination_candidate)
            """
        params = {
            "destination_embedding": destination_embedding,
            "user_id": user_id,
            "threshold": threshold,
        }

        result = self.graph.query(cypher, params=params)
        return result



import sqlite3
import uuid
import threading


class SQLiteManager:
    def __init__(self, db_path=":memory:"):
        self.connection = sqlite3.connect(db_path, check_same_thread=False)
        self._lock = threading.Lock()
        self._migrate_history_table()
        self._create_history_table()

    def _migrate_history_table(self):
        with self._lock:
            with self.connection:
                cursor = self.connection.cursor()

                cursor.execute("SELECT name FROM sqlite_master WHERE type='table' AND name='history'")
                table_exists = cursor.fetchone() is not None

                if table_exists:
                    # Get the current schema of the history table
                    cursor.execute("PRAGMA table_info(history)")
                    current_schema = {row[1]: row[2] for row in cursor.fetchall()}

                    # Define the expected schema
                    expected_schema = {
                        "id": "TEXT",
                        "memory_id": "TEXT",
                        "old_memory": "TEXT",
                        "new_memory": "TEXT",
                        "new_value": "TEXT",
                        "event": "TEXT",
                        "created_at": "DATETIME",
                        "updated_at": "DATETIME",
                        "is_deleted": "INTEGER",
                    }

                    # Check if the schemas are the same
                    if current_schema != expected_schema:
                        # Rename the old table
                        cursor.execute("ALTER TABLE history RENAME TO old_history")

                        cursor.execute(
                            """
                            CREATE TABLE IF NOT EXISTS history (
                                id TEXT PRIMARY KEY,
                                memory_id TEXT,
                                old_memory TEXT,
                                new_memory TEXT,
                                new_value TEXT,
                                event TEXT,
                                created_at DATETIME,
                                updated_at DATETIME,
                                is_deleted INTEGER
                            )
                        """
                        )

                        # Copy data from the old table to the new table
                        cursor.execute(
                            """
                            INSERT INTO history (id, memory_id, old_memory, new_memory, new_value, event, created_at, updated_at, is_deleted)
                            SELECT id, memory_id, prev_value, new_value, new_value, event, timestamp, timestamp, is_deleted
                            FROM old_history
                        """  # noqa: E501
                        )

                        cursor.execute("DROP TABLE old_history")

                        self.connection.commit()

    def _create_history_table(self):
        with self._lock:
            with self.connection:
                self.connection.execute(
                    """
                    CREATE TABLE IF NOT EXISTS history (
                        id TEXT PRIMARY KEY,
                        memory_id TEXT,
                        old_memory TEXT,
                        new_memory TEXT,
                        new_value TEXT,
                        event TEXT,
                        created_at DATETIME,
                        updated_at DATETIME,
                        is_deleted INTEGER
                    )
                """
                )

    def add_history(
        self,
        memory_id,
        old_memory,
        new_memory,
        event,
        created_at=None,
        updated_at=None,
        is_deleted=0,
    ):
        with self._lock:
            with self.connection:
                self.connection.execute(
                    """
                    INSERT INTO history (id, memory_id, old_memory, new_memory, event, created_at, updated_at, is_deleted)
                    VALUES (?, ?, ?, ?, ?, ?, ?, ?)
                """,
                    (
                        str(uuid.uuid4()),
                        memory_id,
                        old_memory,
                        new_memory,
                        event,
                        created_at,
                        updated_at,
                        is_deleted,
                    ),
                )

    def get_history(self, memory_id):
        with self._lock:
            cursor = self.connection.execute(
                """
                SELECT id, memory_id, old_memory, new_memory, event, created_at, updated_at
                FROM history
                WHERE memory_id = ?
                ORDER BY updated_at ASC
            """,
                (memory_id,),
            )
            rows = cursor.fetchall()
            return [
                {
                    "id": row[0],
                    "memory_id": row[1],
                    "old_memory": row[2],
                    "new_memory": row[3],
                    "event": row[4],
                    "created_at": row[5],
                    "updated_at": row[6],
                }
                for row in rows
            ]

    def reset(self):
        with self._lock:
            with self.connection:
                self.connection.execute("DROP TABLE IF EXISTS history")
                self._create_history_table()



import os
from typing import Dict, List, Optional

try:
    import google.generativeai as genai
    from google.generativeai import GenerativeModel, protos
    from google.generativeai.types import content_types
except ImportError:
    raise ImportError(
        "The 'google-generativeai' library is required. Please install it using 'pip install google-generativeai'."
    )

from mem0llama.configs.llms.base import BaseLlmConfig
from mem0llama.llms.base import LLMBase


class GeminiLLM(LLMBase):
    def __init__(self, config: Optional[BaseLlmConfig] = None):
        super().__init__(config)

        if not self.config.model:
            self.config.model = "gemini-1.5-flash-latest"

        api_key = self.config.api_key or os.getenv("GEMINI_API_KEY")
        genai.configure(api_key=api_key)
        self.client = GenerativeModel(model_name=self.config.model)

    def _parse_response(self, response, tools):
        """
        Process the response based on whether tools are used or not.

        Args:
            response: The raw response from API.
            tools: The list of tools provided in the request.

        Returns:
            str or dict: The processed response.
        """
        if tools:
            processed_response = {
                "content": (content if (content := response.candidates[0].content.parts[0].text) else None),
                "tool_calls": [],
            }

            for part in response.candidates[0].content.parts:
                if fn := part.function_call:
                    if isinstance(fn, protos.FunctionCall):
                        fn_call = type(fn).to_dict(fn)
                        processed_response["tool_calls"].append({"name": fn_call["name"], "arguments": fn_call["args"]})
                        continue
                    processed_response["tool_calls"].append({"name": fn.name, "arguments": fn.args})

            return processed_response
        else:
            return response.candidates[0].content.parts[0].text

    def _reformat_messages(self, messages: List[Dict[str, str]]):
        """
        Reformat messages for Gemini.

        Args:
            messages: The list of messages provided in the request.

        Returns:
            list: The list of messages in the required format.
        """
        new_messages = []

        for message in messages:
            if message["role"] == "system":
                content = "THIS IS A SYSTEM PROMPT. YOU MUST OBEY THIS: " + message["content"]

            else:
                content = message["content"]

            new_messages.append(
                {
                    "parts": content,
                    "role": "model" if message["role"] == "model" else "user",
                }
            )

        return new_messages

    def _reformat_tools(self, tools: Optional[List[Dict]]):
        """
        Reformat tools for Gemini.

        Args:
            tools: The list of tools provided in the request.

        Returns:
            list: The list of tools in the required format.
        """

        def remove_additional_properties(data):
            """Recursively removes 'additionalProperties' from nested dictionaries."""

            if isinstance(data, dict):
                filtered_dict = {
                    key: remove_additional_properties(value)
                    for key, value in data.items()
                    if not (key == "additionalProperties")
                }
                return filtered_dict
            else:
                return data

        new_tools = []
        if tools:
            for tool in tools:
                func = tool["function"].copy()
                new_tools.append({"function_declarations": [remove_additional_properties(func)]})

            # TODO: temporarily ignore it to pass tests, will come back to update according to standards later.
            # return content_types.to_function_library(new_tools)

            return new_tools
        else:
            return None

    def generate_response(
        self,
        messages: List[Dict[str, str]],
        response_format=None,
        tools: Optional[List[Dict]] = None,
        tool_choice: str = "auto",
    ):
        """
        Generate a response based on the given messages using Gemini.

        Args:
            messages (list): List of message dicts containing 'role' and 'content'.
            response_format (str or object, optional): Format for the response. Defaults to "text".
            tools (list, optional): List of tools that the model can call. Defaults to None.
            tool_choice (str, optional): Tool choice method. Defaults to "auto".

        Returns:
            str: The generated response.
        """

        params = {
            "temperature": self.config.temperature,
            "max_output_tokens": self.config.max_tokens,
            "top_p": self.config.top_p,
        }

        if response_format is not None and response_format["type"] == "json_object":
            params["response_mime_type"] = "application/json"
            if "schema" in response_format:
                params["response_schema"] = response_format["schema"]
        if tool_choice:
            tool_config = content_types.to_tool_config(
                {
                    "function_calling_config": {
                        "mode": tool_choice,
                        "allowed_function_names": (
                            [tool["function"]["name"] for tool in tools] if tool_choice == "any" else None
                        ),
                    }
                }
            )

        response = self.client.generate_content(
            contents=self._reformat_messages(messages),
            tools=self._reformat_tools(tools),
            generation_config=genai.GenerationConfig(**params),
            tool_config=tool_config,
        )

        return self._parse_response(response, tools)



from typing import Any, Dict, Optional
from enum import Enum

from pydantic import BaseModel, Field, model_validator


class IndexMethod(str, Enum):
    AUTO = "auto"
    HNSW = "hnsw"
    IVFFLAT = "ivfflat"


class IndexMeasure(str, Enum):
    COSINE = "cosine_distance"
    L2 = "l2_distance"
    L1 = "l1_distance"
    MAX_INNER_PRODUCT = "max_inner_product"


class SupabaseConfig(BaseModel):
    connection_string: str = Field(..., description="PostgreSQL connection string")
    collection_name: str = Field("mem0", description="Name for the vector collection")
    embedding_model_dims: Optional[int] = Field(1536, description="Dimensions of the embedding model")
    index_method: Optional[IndexMethod] = Field(IndexMethod.AUTO, description="Index method to use")
    index_measure: Optional[IndexMeasure] = Field(IndexMeasure.COSINE, description="Distance measure to use")

    @model_validator(mode="before")
    def check_connection_string(cls, values):
        conn_str = values.get("connection_string")
        if not conn_str or not conn_str.startswith("postgresql://"):
            raise ValueError("A valid PostgreSQL connection string must be provided")
        return values

    @model_validator(mode="before")
    @classmethod
    def validate_extra_fields(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        allowed_fields = set(cls.model_fields.keys())
        input_fields = set(values.keys())
        extra_fields = input_fields - allowed_fields
        if extra_fields:
            raise ValueError(
                f"Extra fields not allowed: {', '.join(extra_fields)}. Please input only the following fields: {', '.join(allowed_fields)}"
            )
        return values



"""
Utilities for formatting LLM prompts and parsing responses for Mem0's graph memory system.
This module provides functions to create structured prompts for Ollama and other LLMs,
and to parse their responses into standardized formats using Pydantic models.
"""

import json
import logging
from typing import Any, Dict, List, Optional, Union

from mem0llama.memory.models import (
    AddMemory,
    AddMemoryOutput,
    DeleteMemory,
    DeleteMemoryOutput,
    Entity,
    EntityExtraction,
    NoopOutput,
    Relationship,
    RelationshipExtraction,
    UpdateMemory,
    UpdateMemoryOutput,
)

logger = logging.getLogger(__name__)


def create_ollama_format_template(model_class):
    """
    Create a JSON schema format template for Ollama based on a Pydantic model.
    
    Args:
        model_class: A Pydantic model class to generate the schema from
        
    Returns:
        dict: A dictionary containing the JSON schema for the Ollama format parameter
    """
    schema = model_class.model_json_schema()
    return {
        "schema": schema
    }


def parse_llm_response(response: str, model_class) -> Any:
    """
    Parse an LLM response into a Pydantic model.
    
    Args:
        response: The raw response string from the LLM
        model_class: The Pydantic model class to parse the response into
        
    Returns:
        An instance of the specified Pydantic model class
    """
    if not response:
        logger.warning("Empty response from LLM")
        return None
    
    # Handle different response formats
    if isinstance(response, dict):
        # If response is already a dict, it might be from Ollama's direct response
        if "message" in response and "content" in response["message"]:
            content = response["message"]["content"]
        elif "content" in response:
            content = response["content"]
        else:
            # It might already be the parsed JSON
            content = response
    else:
        # Assume it's a string
        content = response
    
    # If content is a string, try to parse it as JSON
    if isinstance(content, str):
        # Extract JSON from markdown code blocks if present
        if "```json" in content:
            start = content.find("```json") + 7
            end = content.find("```", start)
            if end > start:
                content = content[start:end].strip()
        elif "```" in content:
            start = content.find("```") + 3
            end = content.find("```", start)
            if end > start:
                content = content[start:end].strip()
        
        # Try to parse as JSON
        try:
            # Try direct model validation first
            try:
                return model_class.model_validate_json(content)
            except Exception as e:
                # If direct validation fails, try parsing to dict first
                parsed_content = json.loads(content)
                return model_class.model_validate(parsed_content)
        except json.JSONDecodeError as e:
            logger.error(f"Failed to parse JSON: {e}")
            logger.debug(f"Raw content: {content}")
            # Return a default instance as fallback
            return model_class()
    
    # If content is already a dict, validate it directly
    try:
        return model_class.model_validate(content)
    except Exception as e:
        logger.error(f"Failed to validate model: {e}")
        logger.debug(f"Content for validation: {content}")
        # Return a default instance as fallback
        return model_class()


def create_entity_extraction_prompt(text: str, user_id: str) -> str:
    """
    Create a prompt for entity extraction.
    
    Args:
        text: The text to extract entities from
        user_id: The user ID for reference
        
    Returns:
        str: A formatted prompt for entity extraction
    """
    return f"""You are a smart assistant who understands entities and their types in a given text.
If user message contains self-reference such as 'I', 'me', 'my' etc. then use {user_id} as the source entity.
Extract all the entities from the text. DO NOT answer the question itself if the given text is a question.

Extract entities in the following JSON format:
{{
  "entities": [
    {{
      "entity": "entity_name",
      "entity_type": "entity_type"
    }}
  ]
}}

Text: {text}"""


def create_relationship_extraction_prompt(text: str, entities: List[str], user_id: str) -> str:
    """
    Create a prompt for relationship extraction.
    
    Args:
        text: The text to extract relationships from
        entities: A list of entity names
        user_id: The user ID for reference
        
    Returns:
        str: A formatted prompt for relationship extraction
    """
    return f"""You are a smart assistant who understands relationships between entities in text.
Establish relationships among the entities based on the provided text.

Rules:
1. Only establish relationships that are explicitly mentioned or strongly implied in the text.
2. If the text mentions that a relationship no longer exists, do not include it.
3. If the user refers to themselves (using 'I', 'me', 'my', etc.), use '{user_id}' as the entity.

List of entities: {entities}

Text: {text}

Return the relationships in the following JSON format:
{{
  "entities": [
    {{
      "source": "source_entity",
      "relationship": "relationship_type",
      "destination": "destination_entity"
    }}
  ]
}}"""


def create_delete_memory_prompt(existing_memories: str, new_text: str, user_id: str) -> str:
    """
    Create a prompt for identifying memories to delete.
    
    Args:
        existing_memories: String representation of existing memories
        new_text: New text that may contradict existing memories
        user_id: The user ID for reference
        
    Returns:
        str: A formatted prompt for memory deletion
    """
    return f"""You are an assistant managing a knowledge graph of memories.
Based on new information, identify which existing relationships should be deleted because they are:
1. Directly contradicted by the new information
2. No longer true or valid based on the new information
3. Superseded by more accurate or updated information

Existing relationships:
{existing_memories}

New information:
{new_text}

If the user refers to themselves (using 'I', 'me', 'my', etc.), use '{user_id}' as the entity.

Return the relationships to delete in the following JSON format:
{{
  "entities": [
    {{
      "source": "source_entity",
      "relationship": "relationship_type",
      "destination": "destination_entity"
    }}
  ]
}}

If no relationships need to be deleted, return an empty list of entities:
{{
  "entities": []
}}"""


def adapt_llm_provider_for_structured_output(llm_provider: str, llm_config: Dict[str, Any]) -> Dict[str, Any]:
    """
    Adapt the LLM configuration based on the provider to support structured outputs.
    
    Args:
        llm_provider: The LLM provider name (e.g., 'ollama', 'openai')
        llm_config: The original LLM configuration
        
    Returns:
        Dict[str, Any]: Updated LLM configuration
    """
    config = llm_config.copy()
    
    if llm_provider.lower() == 'ollama':
        # For Ollama, we need to handle format differently since BaseLlmConfig doesn't accept it
        # We'll set an environment variable or use a different approach during LLM initialization
        # For now, just ensure temperature is appropriate
        if 'temperature' not in config or config['temperature'] > 0.2:
            config['temperature'] = 0.1
            
        # Add a note in the log about format parameter
        logger.info("For Ollama structured output, format=json should be set during LLM initialization")
    
    elif llm_provider.lower() in ['openai', 'azure_openai']:
        # OpenAI already has good structured output capabilities
        if 'response_format' not in config:
            config['response_format'] = {'type': 'json_object'}
        
        # Ensure temperature is appropriate for structured tasks
        if 'temperature' not in config or config['temperature'] > 0.3:
            config['temperature'] = 0.2
    
    return config



import os
from typing import Dict, List, Optional

try:
    import anthropic
except ImportError:
    raise ImportError("The 'anthropic' library is required. Please install it using 'pip install anthropic'.")

from mem0llama.configs.llms.base import BaseLlmConfig
from mem0llama.llms.base import LLMBase


class AnthropicLLM(LLMBase):
    def __init__(self, config: Optional[BaseLlmConfig] = None):
        super().__init__(config)

        if not self.config.model:
            self.config.model = "claude-3-5-sonnet-20240620"

        api_key = self.config.api_key or os.getenv("ANTHROPIC_API_KEY")
        self.client = anthropic.Anthropic(api_key=api_key)

    def generate_response(
        self,
        messages: List[Dict[str, str]],
        response_format=None,
        tools: Optional[List[Dict]] = None,
        tool_choice: str = "auto",
    ):
        """
        Generate a response based on the given messages using Anthropic.

        Args:
            messages (list): List of message dicts containing 'role' and 'content'.
            response_format (str or object, optional): Format of the response. Defaults to "text".
            tools (list, optional): List of tools that the model can call. Defaults to None.
            tool_choice (str, optional): Tool choice method. Defaults to "auto".

        Returns:
            str: The generated response.
        """
        # Separate system message from other messages
        system_message = ""
        filtered_messages = []
        for message in messages:
            if message["role"] == "system":
                system_message = message["content"]
            else:
                filtered_messages.append(message)

        params = {
            "model": self.config.model,
            "messages": filtered_messages,
            "system": system_message,
            "temperature": self.config.temperature,
            "max_tokens": self.config.max_tokens,
            "top_p": self.config.top_p,
        }
        if tools:  # TODO: Remove tools if no issues found with new memory addition logic
            params["tools"] = tools
            params["tool_choice"] = tool_choice

        response = self.client.messages.create(**params)
        return response.content[0].text



import logging
import traceback
import uuid
from typing import Any, Dict, List, Optional, Tuple

import google.api_core.exceptions
from google.cloud import aiplatform, aiplatform_v1
from google.cloud.aiplatform.matching_engine.matching_engine_index_endpoint import (
    Namespace,
)
from google.oauth2 import service_account
from langchain.schema import Document
from pydantic import BaseModel

from mem0llama.configs.vector_stores.vertex_ai_vector_search import (
    GoogleMatchingEngineConfig,
)
from mem0llama.vector_stores.base import VectorStoreBase

# Configure logging
logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


class OutputData(BaseModel):
    id: Optional[str]  # memory id
    score: Optional[float]  # distance
    payload: Optional[Dict]  # metadata


class GoogleMatchingEngine(VectorStoreBase):
    def __init__(self, **kwargs):
        """Initialize Google Matching Engine client."""
        logger.debug("Initializing Google Matching Engine with kwargs: %s", kwargs)
        
        # If collection_name is passed, use it as deployment_index_id if deployment_index_id is not provided
        if 'collection_name' in kwargs and 'deployment_index_id' not in kwargs:
            kwargs['deployment_index_id'] = kwargs['collection_name']
            logger.debug("Using collection_name as deployment_index_id: %s", kwargs['deployment_index_id'])
        elif 'deployment_index_id' in kwargs and 'collection_name' not in kwargs:
            kwargs['collection_name'] = kwargs['deployment_index_id']
            logger.debug("Using deployment_index_id as collection_name: %s", kwargs['collection_name'])
        
        try:
            config = GoogleMatchingEngineConfig(**kwargs)
            logger.debug("Config created: %s", config.model_dump())
            logger.debug("Config collection_name: %s", getattr(config, 'collection_name', None))
        except Exception as e:
            logger.error("Failed to validate config: %s", str(e))
            raise

        self.project_id = config.project_id
        self.project_number = config.project_number
        self.region = config.region
        self.endpoint_id = config.endpoint_id
        self.index_id = config.index_id  # The actual index ID
        self.deployment_index_id = config.deployment_index_id  # The deployment-specific ID
        self.collection_name = config.collection_name
        self.vector_search_api_endpoint = config.vector_search_api_endpoint
        
        logger.debug("Using project=%s, location=%s", self.project_id, self.region)
        
        # Initialize Vertex AI with credentials if provided
        init_args = {
            "project": self.project_id,
            "location": self.region,
        }
        if hasattr(config, 'credentials_path') and config.credentials_path:
            logger.debug("Using credentials from: %s", config.credentials_path)
            credentials = service_account.Credentials.from_service_account_file(
                config.credentials_path
            )
            init_args["credentials"] = credentials
        
        try:
            aiplatform.init(**init_args)
            logger.debug("Vertex AI initialized successfully")
        except Exception as e:
            logger.error("Failed to initialize Vertex AI: %s", str(e))
            raise
        
        try:
            # Format the index path properly using the configured index_id
            index_path = f"projects/{self.project_number}/locations/{self.region}/indexes/{self.index_id}"
            logger.debug("Initializing index with path: %s", index_path)
            self.index = aiplatform.MatchingEngineIndex(index_name=index_path)
            logger.debug("Index initialized successfully")
            
            # Format the endpoint name properly
            endpoint_name = self.endpoint_id
            logger.debug("Initializing endpoint with name: %s", endpoint_name)
            self.index_endpoint = aiplatform.MatchingEngineIndexEndpoint(
                index_endpoint_name=endpoint_name
            )
            logger.debug("Endpoint initialized successfully")
        except Exception as e:
            logger.error("Failed to initialize Matching Engine components: %s", str(e))
            raise ValueError(f"Invalid configuration: {str(e)}")

    def _parse_output(self, data: Dict) -> List[OutputData]:
        """
        Parse the output data.
        Args:
            data (Dict): Output data.
        Returns:
            List[OutputData]: Parsed output data.
        """
        results = data.get("nearestNeighbors", {}).get("neighbors", [])
        output_data = []
        for result in results:
            output_data.append(
                OutputData(
                    id=result.get("datapoint").get("datapointId"),
                    score=result.get("distance"),
                    payload=result.get("datapoint").get("metadata"),
                )
            )
        return output_data

    def _create_restriction(self, key: str, value: Any) -> aiplatform_v1.types.index.IndexDatapoint.Restriction:
        """Create a restriction object for the Matching Engine index.
        
        Args:
            key: The namespace/key for the restriction
            value: The value to restrict on
        
        Returns:
            Restriction object for the index
        """
        str_value = str(value) if value is not None else ""
        return aiplatform_v1.types.index.IndexDatapoint.Restriction(
            namespace=key,
            allow_list=[str_value]
        )

    def _create_datapoint(
        self, 
        vector_id: str, 
        vector: List[float], 
        payload: Optional[Dict] = None
    ) -> aiplatform_v1.types.index.IndexDatapoint:
        """Create a datapoint object for the Matching Engine index.
        
        Args:
            vector_id: The ID for the datapoint
            vector: The vector to store
            payload: Optional metadata to store with the vector
        
        Returns:
            IndexDatapoint object
        """
        restrictions = []
        if payload:
            restrictions = [
                self._create_restriction(key, value) 
                for key, value in payload.items()
            ]
        
        return aiplatform_v1.types.index.IndexDatapoint(
            datapoint_id=vector_id,
            feature_vector=vector,
            restricts=restrictions
        )

    def insert(
        self,
        vectors: List[list],
        payloads: Optional[List[Dict]] = None,
        ids: Optional[List[str]] = None,
    ) -> None:
        """Insert vectors into the Matching Engine index.
        
        Args:
            vectors: List of vectors to insert
            payloads: Optional list of metadata dictionaries
            ids: Optional list of IDs for the vectors
            
        Raises:
            ValueError: If vectors is empty or lengths don't match
            GoogleAPIError: If the API call fails
        """
        if not vectors:
            raise ValueError("No vectors provided for insertion")
        
        if payloads and len(payloads) != len(vectors):
            raise ValueError(f"Number of payloads ({len(payloads)}) does not match number of vectors ({len(vectors)})")
        
        if ids and len(ids) != len(vectors):
            raise ValueError(f"Number of ids ({len(ids)}) does not match number of vectors ({len(vectors)})")
        
        logger.debug("Starting insert of %d vectors", len(vectors))
        
        try:
            datapoints = [
                self._create_datapoint(
                    vector_id=ids[i] if ids else str(uuid.uuid4()),
                    vector=vector,
                    payload=payloads[i] if payloads and i < len(payloads) else None
                )
                for i, vector in enumerate(vectors)
            ]
            
            logger.debug("Created %d datapoints", len(datapoints))
            self.index.upsert_datapoints(datapoints=datapoints)
            logger.debug("Successfully inserted datapoints")
            
        except google.api_core.exceptions.GoogleAPIError as e:
            logger.error("Failed to insert vectors: %s", str(e))
            raise
        except Exception as e:
            logger.error("Unexpected error during insert: %s", str(e))
            logger.error("Stack trace: %s", traceback.format_exc())
            raise


    def search(self, query: List[float], limit: int = 5, filters: Optional[Dict] = None) -> List[OutputData]:
        """
        Search for similar vectors.
        Args:
            query (List[float]): Query vector.
            limit (int, optional): Number of results to return. Defaults to 5.
            filters (Optional[Dict], optional): Filters to apply to the search. Defaults to None.
        Returns:
            List[OutputData]: Search results (unwrapped)
        """
        logger.debug("Starting search")
        logger.debug("Query type: %s, length: %d", type(query), len(query))
        logger.debug("Limit: %d, Filters: %s", limit, filters)
        
        try:
            filter_namespaces = []
            if filters:
                logger.debug("Processing filters")
                for key, value in filters.items():
                    logger.debug("Processing filter %s=%s (type=%s)", key, value, type(value))
                    if isinstance(value, (str, int, float)):
                        logger.debug("Adding simple filter for %s", key)
                        filter_namespaces.append(
                            Namespace(key, [str(value)], [])
                        )
                    elif isinstance(value, dict):
                        logger.debug("Adding complex filter for %s", key)
                        includes = value.get('include', [])
                        excludes = value.get('exclude', [])
                        filter_namespaces.append(
                            Namespace(key, includes, excludes)
                        )
            
            logger.debug("Final filter_namespaces: %s", filter_namespaces)
            
            response = self.index_endpoint.find_neighbors(
                deployed_index_id=self.deployment_index_id,
                queries=[query],
                num_neighbors=limit,
                filter=filter_namespaces if filter_namespaces else None,
                return_full_datapoint=True
            )
            
            if not response or len(response) == 0 or len(response[0]) == 0:
                logger.debug("No results found")
                return []
            
            results = []
            for neighbor in response[0]:
                logger.debug("Processing neighbor - id: %s, distance: %s", 
                            neighbor.id, neighbor.distance)
                
                payload = {}
                if hasattr(neighbor, 'restricts'):
                    logger.debug("Processing restricts")
                    for restrict in neighbor.restricts:
                        if (hasattr(restrict, 'name') and 
                            hasattr(restrict, 'allow_tokens') and 
                            restrict.allow_tokens):
                            logger.debug("Adding %s: %s", restrict.name, restrict.allow_tokens[0])
                            payload[restrict.name] = restrict.allow_tokens[0]
                
                output_data = OutputData(
                    id=neighbor.id,
                    score=neighbor.distance,
                    payload=payload
                )
                results.append(output_data)
            
            logger.debug("Returning %d results", len(results))
            return results

        except Exception as e:
            logger.error("Error occurred: %s", str(e))
            logger.error("Error type: %s", type(e))
            logger.error("Stack trace: %s", traceback.format_exc())
            raise


    def delete(self, vector_id: Optional[str] = None, ids: Optional[List[str]] = None) -> bool:
        """
        Delete vectors from the Matching Engine index.
        Args:
            vector_id (Optional[str]): Single ID to delete (for backward compatibility)
            ids (Optional[List[str]]): List of IDs of vectors to delete
        Returns:
            bool: True if vectors were deleted successfully or already deleted, False if error
        """
        logger.debug("Starting delete, vector_id: %s, ids: %s", vector_id, ids)
        try:
            # Handle both single vector_id and list of ids
            if vector_id:
                datapoint_ids = [vector_id]
            elif ids:
                datapoint_ids = ids
            else:
                raise ValueError("Either vector_id or ids must be provided")

            logger.debug("Deleting ids: %s", datapoint_ids)
            try:
                self.index.remove_datapoints(datapoint_ids=datapoint_ids)
                logger.debug("Delete completed successfully")
                return True
            except google.api_core.exceptions.NotFound:
                # If the datapoint is already deleted, consider it a success
                logger.debug("Datapoint already deleted")
                return True
            except google.api_core.exceptions.PermissionDenied as e:
                logger.error("Permission denied: %s", str(e))
                return False
            except google.api_core.exceptions.InvalidArgument as e:
                logger.error("Invalid argument: %s", str(e))
                return False
            
        except Exception as e:
            logger.error("Error occurred: %s", str(e))
            logger.error("Error type: %s", type(e))
            logger.error("Stack trace: %s", traceback.format_exc())
            return False


    def update(
        self,
        vector_id: str,
        vector: Optional[List[float]] = None,
        payload: Optional[Dict] = None,
    ) -> bool:
        """Update a vector and its payload.
        
        Args:
            vector_id: ID of the vector to update
            vector: Optional new vector values
            payload: Optional new metadata payload
            
        Returns:
            bool: True if update was successful
            
        Raises:
            ValueError: If neither vector nor payload is provided
            GoogleAPIError: If the API call fails
        """
        logger.debug("Starting update for vector_id: %s", vector_id)
        
        if vector is None and payload is None:
            raise ValueError("Either vector or payload must be provided for update")
        
        # First check if the vector exists
        try:
            existing = self.get(vector_id)
            if existing is None:
                logger.error("Vector ID not found: %s", vector_id)
                return False
            
            datapoint = self._create_datapoint(
                vector_id=vector_id,
                vector=vector if vector is not None else [],
                payload=payload
            )
            
            logger.debug("Upserting datapoint: %s", datapoint)
            self.index.upsert_datapoints(datapoints=[datapoint])
            logger.debug("Update completed successfully")
            return True
            
        except google.api_core.exceptions.GoogleAPIError as e:
            logger.error("API error during update: %s", str(e))
            return False
        except Exception as e:
            logger.error("Unexpected error during update: %s", str(e))
            logger.error("Stack trace: %s", traceback.format_exc())
            raise


    def get(self, vector_id: str) -> Optional[OutputData]:
        """
        Retrieve a vector by ID.
        Args:
            vector_id (str): ID of the vector to retrieve.
        Returns:
            Optional[OutputData]: Retrieved vector or None if not found.
        """
        logger.debug("Starting get for vector_id: %s", vector_id)
        
        try:
            if not self.vector_search_api_endpoint:
                raise ValueError("vector_search_api_endpoint is required for get operation")

            vector_search_client = aiplatform_v1.MatchServiceClient(
                client_options={
                    "api_endpoint": self.vector_search_api_endpoint
                },
            )
            datapoint = aiplatform_v1.IndexDatapoint(
                datapoint_id=vector_id
            )

            query = aiplatform_v1.FindNeighborsRequest.Query(
                datapoint=datapoint,
                neighbor_count=1
            )
            request = aiplatform_v1.FindNeighborsRequest(
                index_endpoint=f"projects/{self.project_number}/locations/{self.region}/indexEndpoints/{self.endpoint_id}",
                deployed_index_id=self.deployment_index_id,
                queries=[query],
                return_full_datapoint=True,
            )

            try:
                response = vector_search_client.find_neighbors(request)
                logger.debug("Got response")
                
                if response and response.nearest_neighbors:
                    nearest = response.nearest_neighbors[0]
                    if nearest.neighbors:
                        neighbor = nearest.neighbors[0]
                        
                        payload = {}
                        if hasattr(neighbor.datapoint, 'restricts'):
                            for restrict in neighbor.datapoint.restricts:
                                if restrict.allow_list:
                                    payload[restrict.namespace] = restrict.allow_list[0]
                
                        return OutputData(
                            id=neighbor.datapoint.datapoint_id,
                            score=neighbor.distance,
                            payload=payload
                        )
                
                logger.debug("No results found")
                return None
                
            except google.api_core.exceptions.NotFound:
                logger.debug("Datapoint not found")
                return None
            except google.api_core.exceptions.PermissionDenied as e:
                logger.error("Permission denied: %s", str(e))
                return None
            
        except Exception as e:
            logger.error("Error occurred: %s", str(e))
            logger.error("Error type: %s", type(e))
            logger.error("Stack trace: %s", traceback.format_exc())
            raise


    def list_cols(self) -> List[str]:
        """
        List all collections (indexes).
        Returns:
            List[str]: List of collection names.
        """
        return [self.deployment_index_id]


    def delete_col(self):
        """
        Delete a collection (index).
        Note: This operation is not supported through the API.
        """
        logger.warning("Delete collection operation is not supported for Google Matching Engine")
        pass


    def col_info(self) -> Dict:
        """
        Get information about a collection (index).
        Returns:
            Dict: Collection information.
        """
        return {
            "index_id": self.index_id,
            "endpoint_id": self.endpoint_id,
            "project_id": self.project_id,
            "region": self.region
        }


    def list(self, filters: Optional[Dict] = None, limit: Optional[int] = None) -> List[List[OutputData]]:
        """List vectors matching the given filters.
        
        Args:
            filters: Optional filters to apply
            limit: Optional maximum number of results to return
            
        Returns:
            List[List[OutputData]]: List of matching vectors wrapped in an extra array
            to match the interface
        """
        logger.debug("Starting list operation")
        logger.debug("Filters: %s", filters)
        logger.debug("Limit: %s", limit)
        
        try:
            # Use a zero vector for the search
            dimension = 768  # This should be configurable based on the model
            zero_vector = [0.0] * dimension
            
            # Use a large limit if none specified
            search_limit = limit if limit is not None else 10000
            
            results = self.search(
                query=zero_vector,
                limit=search_limit,
                filters=filters
            )
            
            logger.debug("Found %d results", len(results))
            return [results]  # Wrap in extra array to match interface
            
        except Exception as e:
            logger.error("Error in list operation: %s", str(e))
            logger.error("Stack trace: %s", traceback.format_exc())
            raise


    def create_col(self, name=None, vector_size=None, distance=None):
        """
        Create a new collection. For Google Matching Engine, collections (indexes) 
        are created through the Google Cloud Console or API separately.
        This method is a no-op since indexes are pre-created.
        
        Args:
            name: Ignored for Google Matching Engine
            vector_size: Ignored for Google Matching Engine
            distance: Ignored for Google Matching Engine
        """
        # Google Matching Engine indexes are created through Google Cloud Console
        # This method is included only to satisfy the abstract base class
        pass


    def add(self, text: str, metadata: Optional[Dict] = None, user_id: Optional[str] = None) -> str:
        logger.debug("Starting add operation")
        logger.debug("Text: %s", text)
        logger.debug("Metadata: %s", metadata)
        logger.debug("User ID: %s", user_id)
        
        try:
            # Generate a unique ID for this entry
            vector_id = str(uuid.uuid4())
            
            # Create the payload with all necessary fields
            payload = {
                "data": text,  # Store the text in the data field
                "user_id": user_id,
                **(metadata or {})
            }
            
            # Get the embedding
            vector = self.embedder.embed_query(text)
            
            # Insert using the insert method
            self.insert(
                vectors=[vector],
                payloads=[payload],
                ids=[vector_id]
            )
            
            return vector_id
            
        except Exception as e:
            logger.error("Error occurred: %s", str(e))
            raise


    def add_texts(
        self,
        texts: List[str],
        metadatas: Optional[List[dict]] = None,
        ids: Optional[List[str]] = None,
    ) -> List[str]:
        """Add texts to the vector store.
        
        Args:
            texts: List of texts to add
            metadatas: Optional list of metadata dicts
            ids: Optional list of IDs to use
            
        Returns:
            List[str]: List of IDs of the added texts
            
        Raises:
            ValueError: If texts is empty or lengths don't match
        """
        if not texts:
            raise ValueError("No texts provided")
        
        if metadatas and len(metadatas) != len(texts):
            raise ValueError(f"Number of metadata items ({len(metadatas)}) does not match number of texts ({len(texts)})")
        
        if ids and len(ids) != len(texts):
            raise ValueError(f"Number of ids ({len(ids)}) does not match number of texts ({len(texts)})")
        
        logger.debug("Starting add_texts operation")
        logger.debug("Number of texts: %d", len(texts))
        logger.debug("Has metadatas: %s", metadatas is not None)
        logger.debug("Has ids: %s", ids is not None)
        
        if ids is None:
            ids = [str(uuid.uuid4()) for _ in texts]
        
        try:
            # Get embeddings
            embeddings = self.embedder.embed_documents(texts)
            
            # Add to store
            self.insert(
                vectors=embeddings,
                payloads=metadatas if metadatas else [{}] * len(texts),
                ids=ids
            )
            return ids
            
        except Exception as e:
            logger.error("Error in add_texts: %s", str(e))
            logger.error("Stack trace: %s", traceback.format_exc())
            raise

    @classmethod
    def from_texts(
        cls,
        texts: List[str],
        embedding: Any,
        metadatas: Optional[List[dict]] = None,
        ids: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> "GoogleMatchingEngine":
        """Create an instance from texts."""
        logger.debug("Creating instance from texts")
        store = cls(**kwargs)
        store.add_texts(texts=texts, metadatas=metadatas, ids=ids)
        return store

    def similarity_search_with_score(
        self,
        query: str,
        k: int = 5,
        filter: Optional[Dict] = None,
    ) -> List[Tuple[Document, float]]:
        """Return documents most similar to query with scores."""
        logger.debug("Starting similarity search with score")
        logger.debug("Query: %s", query)
        logger.debug("k: %d", k)
        logger.debug("Filter: %s", filter)
        
        embedding = self.embedder.embed_query(query)
        results = self.search(query=embedding, limit=k, filters=filter)
        
        docs_and_scores = [
            (
                Document(
                    page_content=result.payload.get("text", ""),
                    metadata=result.payload
                ),
                result.score
            )
            for result in results
        ]
        logger.debug("Found %d results", len(docs_and_scores))
        return docs_and_scores

    def similarity_search(
        self,
        query: str,
        k: int = 5,
        filter: Optional[Dict] = None,
    ) -> List[Document]:
        """Return documents most similar to query."""
        logger.debug("Starting similarity search")
        docs_and_scores = self.similarity_search_with_score(query, k, filter)
        return [doc for doc, _ in docs_and_scores]




import json
import os
from typing import Dict, List, Optional

from openai import AzureOpenAI

from mem0llama.configs.llms.base import BaseLlmConfig
from mem0llama.llms.base import LLMBase


class AzureOpenAIStructuredLLM(LLMBase):
    def __init__(self, config: Optional[BaseLlmConfig] = None):
        super().__init__(config)

        # Model name should match the custom deployment name chosen for it.
        if not self.config.model:
            self.config.model = "gpt-4o-2024-08-06"

        api_key = os.getenv("LLM_AZURE_OPENAI_API_KEY") or self.config.azure_kwargs.api_key
        azure_deployment = os.getenv("LLM_AZURE_DEPLOYMENT") or self.config.azure_kwargs.azure_deployment
        azure_endpoint = os.getenv("LLM_AZURE_ENDPOINT") or self.config.azure_kwargs.azure_endpoint
        api_version = os.getenv("LLM_AZURE_API_VERSION") or self.config.azure_kwargs.api_version
        default_headers = self.config.azure_kwargs.default_headers

        # Can display a warning if API version is of model and api-version
        self.client = AzureOpenAI(
            azure_deployment=azure_deployment,
            azure_endpoint=azure_endpoint,
            api_version=api_version,
            api_key=api_key,
            http_client=self.config.http_client,
            default_headers=default_headers,
        )

    def _parse_response(self, response, tools):
        """
        Process the response based on whether tools are used or not.

        Args:
            response: The raw response from API.
            tools: The list of tools provided in the request.

        Returns:
            str or dict: The processed response.
        """
        if tools:
            processed_response = {
                "content": response.choices[0].message.content,
                "tool_calls": [],
            }

            if response.choices[0].message.tool_calls:
                for tool_call in response.choices[0].message.tool_calls:
                    processed_response["tool_calls"].append(
                        {
                            "name": tool_call.function.name,
                            "arguments": json.loads(tool_call.function.arguments),
                        }
                    )

            return processed_response
        else:
            return response.choices[0].message.content

    def generate_response(
        self,
        messages: List[Dict[str, str]],
        response_format=None,
        tools: Optional[List[Dict]] = None,
        tool_choice: str = "auto",
    ):
        """
        Generate a response based on the given messages using Azure OpenAI.

        Args:
            messages (list): List of message dicts containing 'role' and 'content'.
            response_format (str or object, optional): Format of the response. Defaults to "text".
            tools (list, optional): List of tools that the model can call. Defaults to None.
            tool_choice (str, optional): Tool choice method. Defaults to "auto".

        Returns:
            str: The generated response.
        """
        params = {
            "model": self.config.model,
            "messages": messages,
            "temperature": self.config.temperature,
            "max_tokens": self.config.max_tokens,
            "top_p": self.config.top_p,
        }
        if response_format:
            params["response_format"] = response_format
        if tools:
            params["tools"] = tools
            params["tool_choice"] = tool_choice

        response = self.client.chat.completions.create(**params)
        return self._parse_response(response, tools)



from typing import Optional

from pydantic import BaseModel, Field


class GoogleMatchingEngineConfig(BaseModel):
    project_id: str = Field(description="Google Cloud project ID")
    project_number: str = Field(description="Google Cloud project number")
    region: str = Field(description="Google Cloud region")
    endpoint_id: str = Field(description="Vertex AI Vector Search endpoint ID")
    index_id: str = Field(description="Vertex AI Vector Search index ID")
    deployment_index_id: str = Field(description="Deployment-specific index ID")
    collection_name: Optional[str] = Field(None, description="Collection name, defaults to index_id")
    credentials_path: Optional[str] = Field(None, description="Path to service account credentials file")
    vector_search_api_endpoint: Optional[str] = Field(None, description="Vector search API endpoint")

    model_config = {
        "extra": "forbid"
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if not self.collection_name:
            self.collection_name = self.index_id

    def model_post_init(self, _context) -> None:
        """Set collection_name to index_id if not provided"""
        if self.collection_name is None:
            self.collection_name = self.index_id


from typing import Optional

from pydantic import BaseModel, Field, field_validator


class EmbedderConfig(BaseModel):
    provider: str = Field(
        description="Provider of the embedding model (e.g., 'ollama', 'openai')",
        default="openai",
    )
    config: Optional[dict] = Field(description="Configuration for the specific embedding model", default={})

    @field_validator("config")
    def validate_config(cls, v, values):
        provider = values.data.get("provider")
        if provider in ["openai", "ollama", "huggingface", "azure_openai", "gemini", "vertexai", "together"]:
            return v
        else:
            raise ValueError(f"Unsupported embedding provider: {provider}")



import importlib.metadata

__version__ = importlib.metadata.version("mem0llama")

from mem0llama.client.main import MemoryClient, AsyncMemoryClient  # noqa
from mem0llama.memory.main import Memory  # noqa



from typing import Literal, Optional

from sentence_transformers import SentenceTransformer

from mem0llama.configs.embeddings.base import BaseEmbedderConfig
from mem0llama.embeddings.base import EmbeddingBase


class HuggingFaceEmbedding(EmbeddingBase):
    def __init__(self, config: Optional[BaseEmbedderConfig] = None):
        super().__init__(config)

        self.config.model = self.config.model or "multi-qa-MiniLM-L6-cos-v1"

        self.model = SentenceTransformer(self.config.model, **self.config.model_kwargs)

        self.config.embedding_dims = self.config.embedding_dims or self.model.get_sentence_embedding_dimension()

    def embed(self, text, memory_action: Optional[Literal["add", "search", "update"]] = None):
        """
        Get the embedding for the given text using Hugging Face.

        Args:
            text (str): The text to embed.
            memory_action (optional): The type of embedding to use. Must be one of "add", "search", or "update". Defaults to None.
        Returns:
            list: The embedding vector.
        """
        return self.model.encode(text, convert_to_numpy=True).tolist()



import re
from mem0llama.configs.prompts import FACT_RETRIEVAL_PROMPT
from mem0llama.llms.openai import OpenAILLM


def get_fact_retrieval_messages(message):
    return FACT_RETRIEVAL_PROMPT, f"Input:\n{message}"


def parse_messages(messages):
    response = ""
    for msg in messages:
        if msg["role"] == "system":
            response += f"system: {msg['content']}\n"
        if msg["role"] == "user":
            response += f"user: {msg['content']}\n"
        if msg["role"] == "assistant":
            response += f"assistant: {msg['content']}\n"
    return response


def format_entities(entities):
    if not entities:
        return ""

    formatted_lines = []
    for entity in entities:
        simplified = f"{entity['source']} -- {entity['relationship']} -- {entity['destination']}"
        formatted_lines.append(simplified)

    return "\n".join(formatted_lines)


def remove_code_blocks(content: str) -> str:
    """
    Removes enclosing code block markers ```[language] and ``` from a given string.

    Remarks:
    - The function uses a regex pattern to match code blocks that may start with ``` followed by an optional language tag (letters or numbers) and end with ```.
    - If a code block is detected, it returns only the inner content, stripping out the markers.
    - If no code block markers are found, the original content is returned as-is.
    """
    pattern = r"^```[a-zA-Z0-9]*\n([\s\S]*?)\n```$"
    match = re.match(pattern, content.strip())
    return match.group(1).strip() if match else content.strip()


def get_image_description(image_url):
    """
    Get the description of the image
    """
    llm = OpenAILLM()
    response = llm.generate_response(
        messages=[
            {
                "role": "user",
                "content": [
                    {
                        "type": "text",
                        "text": "A user is providing an image. Provide a high level description of the image and do not include any additional text.",
                    },
                    {"type": "image_url", "image_url": {"url": image_url}},
                ],
            },
        ]
    )
    return response


def parse_vision_messages(messages):
    """
    Parse the vision messages from the messages
    """
    returned_messages = []
    for msg in messages:
        if msg["role"] != "system":
            if not isinstance(msg["content"], str) and msg["content"]["type"] == "image_url":
                image_url = msg["content"]["image_url"]["url"]
                try:
                    description = get_image_description(image_url)
                    msg["content"]["text"] = description
                    returned_messages.append({"role": msg["role"], "content": description})
                except Exception:
                    raise Exception(f"Error while downloading {image_url}.")
            else:
                returned_messages.append(msg)
        else:
            returned_messages.append(msg)
    return returned_messages
